@@ -1,8 +1,12 @@
+// SPDX-FileCopyrightText: 2023 Greenbone AG
+//
+// SPDX-License-Identifier: GPL-2.0-or-later
+
 use crate::{
     error::SyntaxError,
     grouping_extension::Grouping,
     lexer::{End, Lexer},
-    token::{Category, Keyword, Token},
+    token::{Category, IdentifierType, Token},
     unclosed_statement, unclosed_token, unexpected_end, unexpected_statement, unexpected_token,
     variable_extension::CommaGroup,
     DeclareScope, Statement,
@@ -12,7 +16,7 @@ pub(crate) trait Keywords {
     /// Parses keywords.
     fn parse_keyword(
         &mut self,
-        keyword: Keyword,
+        keyword: IdentifierType,
         token: Token,
     ) -> Result<(End, Statement), SyntaxError>;
 }
@@ -35,28 +39,26 @@ impl<'a> Lexer<'a> {
     fn parse_if(&mut self) -> Result<(End, Statement), SyntaxError> {
         let token = self.token().ok_or_else(|| unexpected_end!("if parsing"))?;
         let condition = match token.category() {
-            Category::LeftParen => self.parse_paren(token),
-            _ => Err(unexpected_token!(token)),
+            Category::LeftParen => self.parse_paren(token.clone()),
+            _ => Err(unexpected_token!(token.clone())),
         }?
         .as_returnable_or_err()?;
-        let (end, body) = self.statement(0, &|cat| cat == Category::Semicolon)?;
+        let (end, body) = self.statement(0, &|cat| cat == &Category::Semicolon)?;
         if end == End::Continue {
             return Err(unclosed_token!(token));
         }
         let r#else: Option<Statement> = {
             match self.peek() {
                 Some(token) => match token.category() {
-                    Category::Identifier(Some(Keyword::Else)) => {
+                    Category::Identifier(IdentifierType::Else) => {
                         self.token();
-                        let (end, stmt) = self.statement(0, &|cat| cat == Category::Semicolon)?;
+                        let (end, stmt) = self.statement(0, &|cat| cat == &Category::Semicolon)?;
                         if end == End::Continue {
                             return Err(unexpected_statement!(stmt));
                         }
                         Some(stmt)
                     }
-                    _ => {
-                        None
-                    }
+                    _ => None,
                 },
                 None => None,
             }
@@ -71,7 +73,7 @@ impl<'a> Lexer<'a> {
         let token = self
             .token()
             .ok_or_else(|| unexpected_end!("expected paren."))?;
-        if token.category() != Category::LeftParen {
+        if token.category() != &Category::LeftParen {
             Err(unexpected_token!(token))
         } else {
             Ok(())
@@ -80,7 +82,7 @@ impl<'a> Lexer<'a> {
 
     fn parse_call_return_params(&mut self) -> Result<Statement, SyntaxError> {
         self.jump_to_left_parenthesis()?;
-        let (end, parameter) = self.statement(0, &|cat| cat == Category::RightParen)?;
+        let (end, parameter) = self.statement(0, &|cat| cat == &Category::RightParen)?;
         let parameter = parameter.as_returnable_or_err()?;
         if end.is_done() {
             Ok(parameter)
@@ -91,7 +93,7 @@ impl<'a> Lexer<'a> {
 
     fn parse_exit(&mut self) -> Result<(End, Statement), SyntaxError> {
         let parameter = self.parse_call_return_params()?;
-        let (_, should_be_semicolon) = self.statement(0, &|cat| cat == Category::Semicolon)?;
+        let (_, should_be_semicolon) = self.statement(0, &|cat| cat == &Category::Semicolon)?;
         if matches!(should_be_semicolon, Statement::NoOp(_)) {
             Ok((
                 End::Done(Category::Semicolon),
@@ -104,12 +106,17 @@ impl<'a> Lexer<'a> {
 
     fn parse_include(&mut self) -> Result<(End, Statement), SyntaxError> {
         let parameter = self.parse_call_return_params()?;
-        match parameter {
-            Statement::Primitive(_) | Statement::Variable(_) | Statement::Array(_, _) => Ok((
-                End::Done(Category::RightParen),
-                Statement::Include(Box::new(parameter)),
-            )),
-            _ => Err(unexpected_statement!(parameter)),
+        let (_, should_be_semicolon) = self.statement(0, &|cat| cat == &Category::Semicolon)?;
+        if matches!(should_be_semicolon, Statement::NoOp(_)) {
+            match parameter {
+                Statement::Primitive(_) | Statement::Variable(_) | Statement::Array(_, _) => Ok((
+                    End::Done(Category::RightParen),
+                    Statement::Include(Box::new(parameter)),
+                )),
+                _ => Err(unexpected_statement!(parameter)),
+            }
+        } else {
+            Err(unexpected_statement!(should_be_semicolon))
         }
     }
 
@@ -117,7 +124,10 @@ impl<'a> Lexer<'a> {
         let id = self
             .token()
             .ok_or_else(|| unexpected_end!("parse_function"))?;
-        if !matches!(id.category(), Category::Identifier(None)) {
+        if !matches!(
+            id.category(),
+            Category::Identifier(IdentifierType::Undefined(_))
+        ) {
             return Err(unexpected_token!(id));
         }
         let paren = self
@@ -155,37 +165,87 @@ impl<'a> Lexer<'a> {
                 ));
             }
         }
-        let (end, parameter) = self.statement(0, &|cat| cat == Category::Semicolon)?;
+        let (end, parameter) = self.statement(0, &|cat| cat == &Category::Semicolon)?;
         let parameter = parameter.as_returnable_or_err()?;
         if let End::Done(cat) = end {
-            Ok((
-                End::Done(cat),
-                Statement::Return(Box::new(parameter)),
-            ))
+            Ok((End::Done(cat), Statement::Return(Box::new(parameter))))
         } else {
             Err(unexpected_end!("exit"))
         }
     }
+    fn parse_continue(&mut self) -> Result<(End, Statement), SyntaxError> {
+        let token = self.peek();
+        if let Some(token) = token {
+            if matches!(token.category(), Category::Semicolon) {
+                self.token();
+                return Ok((End::Done(Category::Semicolon), Statement::Continue));
+            } else {
+                return Err(unexpected_token!(token));
+            }
+        }
+        Err(unexpected_end!("exit"))
+    }
+    fn parse_break(&mut self) -> Result<(End, Statement), SyntaxError> {
+        let token = self.peek();
+        if let Some(token) = token {
+            if matches!(token.category(), Category::Semicolon) {
+                self.token();
+                return Ok((End::Done(Category::Semicolon), Statement::Break));
+            } else {
+                return Err(unexpected_token!(token));
+            }
+        }
+        Err(unexpected_end!("exit"))
+    }
+
+    fn map_syntax_error_to_unclosed_left_paren(e: SyntaxError) -> SyntaxError {
+        match e.kind() {
+            crate::ErrorKind::UnexpectedToken(k) => unclosed_token!(Token {
+                category: Category::LeftParen,
+                position: k.position
+            }),
+            _ => e,
+        }
+    }
+
     fn parse_for(&mut self) -> Result<(End, Statement), SyntaxError> {
         self.jump_to_left_parenthesis()?;
-        let (end, assignment) = self.statement(0, &|c| c == Category::Semicolon)?;
-        if !matches!(assignment, Statement::Assign(_, _, _, _)) {
+        let (end, assignment) = self.statement(0, &|c| c == &Category::Semicolon)?;
+        if !matches!(
+            assignment,
+            Statement::Assign(_, _, _, _) | Statement::NoOp(_)
+        ) {
             return Err(unexpected_statement!(assignment));
         }
         if end == End::Continue {
             return Err(unclosed_statement!(assignment));
         }
         // `for (i = 0; i < 10; i++) display("hi");`
-        let (end, condition) = self.statement(0, &|c| c == Category::Semicolon)?;
+        let (end, condition) = self.statement(0, &|c| c == &Category::Semicolon)?;
         let condition = condition.as_returnable_or_err()?;
         if end == End::Continue {
             return Err(unclosed_statement!(condition));
         }
-        let (end, update) = self.statement(0, &|c| c == Category::RightParen)?;
-        if end == End::Continue {
-            return Err(unclosed_statement!(update));
+        let (end, update) = match self.peek() {
+            // no update statement provided
+            Some(Token {
+                category: Category::RightParen,
+                position: _,
+            }) => {
+                self.token();
+                (End::Done(Category::RightParen), Statement::NoOp(None))
+            }
+            _ => self
+                .statement(0, &|c| c == &Category::RightParen)
+                .map_err(Self::map_syntax_error_to_unclosed_left_paren)?,
+        };
+        if !matches!(end, End::Done(Category::RightParen)) {
+            return Err(unclosed_token!(Token {
+                category: Category::LeftParen,
+                position: update.as_token().map_or_else(|| (0, 0), |t| t.position)
+            }));
         }
-        let (end, body) = self.statement(0, &|c| c == Category::Semicolon)?;
+        let (end, body) = self.statement(0, &|c| c == &Category::Semicolon)?;
         match end {
             End::Done(cat) => Ok((
                 End::Done(cat),
@@ -202,12 +262,17 @@ impl<'a> Lexer<'a> {
 
     fn parse_while(&mut self, token: Token) -> Result<(End, Statement), SyntaxError> {
         self.jump_to_left_parenthesis()?;
-        let (end, condition) = self.statement(0, &|c| c == Category::RightParen)?;
-        if !end {
-            return Err(unclosed_token!(token));
+        let (end, condition) = self
+            .statement(0, &|c| c == &Category::RightParen)
+            .map_err(Self::map_syntax_error_to_unclosed_left_paren)?;
+        if !matches!(end, End::Done(Category::RightParen)) {
+            return Err(unclosed_token!(Token {
+                category: Category::LeftParen,
+                position: condition.as_token().map_or_else(|| (0, 0), |t| t.position)
+            }));
         }
         let condition = condition.as_returnable_or_err()?;
-        let (end, body) = self.statement(0, &|c| c == Category::Semicolon)?;
+        let (end, body) = self.statement(0, &|c| c == &Category::Semicolon)?;
         if !end {
             return Err(unclosed_token!(token));
         }
@@ -217,7 +282,7 @@ impl<'a> Lexer<'a> {
         ))
     }
     fn parse_repeat(&mut self, token: Token) -> Result<(End, Statement), SyntaxError> {
-        let (end, body) = self.statement(0, &|c| c == Category::Semicolon)?;
+        let (end, body) = self.statement(0, &|c| c == &Category::Semicolon)?;
 
         if !end {
             return Err(unclosed_token!(token));
@@ -225,8 +290,8 @@ impl<'a> Lexer<'a> {
         let until: Statement = {
             match self.token() {
                 Some(token) => match token.category() {
-                    Category::Identifier(Some(Keyword::Until)) => {
-                        let (end, stmt) = self.statement(0, &|cat| cat == Category::Semicolon)?;
+                    Category::Identifier(IdentifierType::Until) => {
+                        let (end, stmt) = self.statement(0, &|cat| cat == &Category::Semicolon)?;
                         if !end {
                             return Err(unclosed_token!(token));
                         }
@@ -248,7 +313,7 @@ impl<'a> Lexer<'a> {
         let variable: Token = {
             match self.token() {
                 Some(token) => match token.category() {
-                    Category::Identifier(None) => Ok(token),
+                    Category::Identifier(IdentifierType::Undefined(_)) => Ok(token),
                     _ => Err(unexpected_token!(token)),
                 },
                 None => Err(unexpected_end!("in foreach")),
@@ -257,11 +322,13 @@ impl<'a> Lexer<'a> {
         let r#in: Statement = {
             let token = self.token().ok_or_else(|| unexpected_end!("in foreach"))?;
             match token.category() {
-                Category::LeftParen => self.parse_paren(token),
+                Category::LeftParen => self
+                    .parse_paren(token.clone())
+                    .map_err(|_| unclosed_token!(token)),
                 _ => Err(unexpected_token!(token)),
             }?
         };
-        let (end, block) = self.statement(0, &|cat| cat == Category::Semicolon)?;
+        let (end, block) = self.statement(0, &|cat| cat == &Category::Semicolon)?;
         if !end {
             Err(unclosed_token!(token))
         } else {
@@ -271,29 +338,23 @@ impl<'a> Lexer<'a> {
             ))
         }
     }
-    fn parse_fct_anon_args(
-        &mut self,
-        keyword: Token,
-    ) -> Result<(End, Statement), SyntaxError> {
+    fn parse_fct_anon_args(&mut self, keyword: Token) -> Result<(End, Statement), SyntaxError> {
         match self.peek() {
             Some(token) => match token.category() {
                 Category::LeftBrace => {
                     self.token();
-                    let (end, lookup) = self.statement(0, &|c| c == Category::RightBrace)?;
+                    let (end, lookup) = self.statement(0, &|c| c == &Category::RightBrace)?;
                     let lookup = lookup.as_returnable_or_err()?;
                     if end == End::Continue {
                         Err(unclosed_token!(token))
                     } else {
-
                         Ok((
                             End::Continue,
                             Statement::Array(keyword, Some(Box::new(lookup))),
                         ))
                     }
                 }
-                _ => {
-                    Ok((End::Continue, Statement::Array(keyword, None)))
-                }
+                _ => Ok((End::Continue, Statement::Array(keyword, None))),
             },
             None => Err(unexpected_end!("in fct_anon_args")),
         }
@@ -303,28 +364,33 @@ impl<'a> Lexer<'a> {
 impl<'a> Keywords for Lexer<'a> {
     fn parse_keyword(
         &mut self,
-        keyword: Keyword,
+        keyword: IdentifierType,
         token: Token,
     ) -> Result<(End, Statement), SyntaxError> {
         match keyword {
-            Keyword::For => self.parse_for(),
-            Keyword::ForEach => self.parse_foreach(token),
-            Keyword::If => self.parse_if(),
-            Keyword::Else => Err(unexpected_token!(token)), // handled in if
-            Keyword::While => self.parse_while(token),
-            Keyword::Repeat => self.parse_repeat(token),
-            Keyword::Until => Err(unexpected_token!(token)), // handled in repeat
-            Keyword::LocalVar => self.parse_declaration(DeclareScope::Local),
-            Keyword::GlobalVar => self.parse_declaration(DeclareScope::Global),
-            Keyword::Null => Ok((End::Continue, Statement::Primitive(token))),
-            Keyword::Return => self.parse_return(),
-            Keyword::Include => self.parse_include(),
-            Keyword::Exit => self.parse_exit(),
-            Keyword::FCTAnonArgs => self.parse_fct_anon_args(token),
-            Keyword::True => Ok((End::Continue, Statement::Primitive(token))),
-            Keyword::False => Ok((End::Continue, Statement::Primitive(token))),
-            Keyword::Function => self.parse_function(token),
-            Keyword::ACT(category) => Ok((End::Continue, Statement::AttackCategory(category))),
+            IdentifierType::For => self.parse_for(),
+            IdentifierType::ForEach => self.parse_foreach(token),
+            IdentifierType::If => self.parse_if(),
+            IdentifierType::Else => Err(unexpected_token!(token)), // handled in if
+            IdentifierType::While => self.parse_while(token),
+            IdentifierType::Repeat => self.parse_repeat(token),
+            IdentifierType::Until => Err(unexpected_token!(token)), // handled in repeat
+            IdentifierType::LocalVar => self.parse_declaration(DeclareScope::Local),
+            IdentifierType::GlobalVar => self.parse_declaration(DeclareScope::Global),
+            IdentifierType::Null => Ok((End::Continue, Statement::Primitive(token))),
+            IdentifierType::Return => self.parse_return(),
+            IdentifierType::Include => self.parse_include(),
+            IdentifierType::Exit => self.parse_exit(),
+            IdentifierType::FCTAnonArgs => self.parse_fct_anon_args(token),
+            IdentifierType::True => Ok((End::Continue, Statement::Primitive(token))),
+            IdentifierType::False => Ok((End::Continue, Statement::Primitive(token))),
+            IdentifierType::Function => self.parse_function(token),
+            IdentifierType::ACT(category) => {
+                Ok((End::Continue, Statement::AttackCategory(category)))
+            }
+            IdentifierType::Undefined(_) => Err(unexpected_token!(token)),
+            IdentifierType::Continue => self.parse_continue(),
+            IdentifierType::Break => self.parse_break(),
         }
     }
 }
@@ -334,10 +400,11 @@ mod test {
 
     use crate::{
         parse,
-        token::{Base, Category, Keyword, StringCategory, Token},
+        token::{Category, IdentifierType, Token},
         AssignOrder, DeclareScope, SyntaxError,
     };
 
+    use crate::IdentifierType::Undefined;
     use crate::Statement::*;
     use crate::TokenCategory::*;
 
@@ -351,28 +418,28 @@ mod test {
             actual,
             If(
                 Box::new(Variable(Token {
-                    category: Identifier(None),
-                    position: (4, 15)
+                    category: Identifier(Undefined("description".to_owned())),
+                    position: (1, 5)
                 })),
                 Box::new(Call(
                     Token {
-                        category: Identifier(None),
-                        position: (17, 27)
+                        category: Identifier(Undefined("script_oid".to_owned())),
+                        position: (1, 18)
                     },
-                    Box::new(Parameter(vec![Primitive(Token {
-                        category: String(StringCategory::Quotable),
-                        position: (29, 30)
-                    })]))
+                    vec![Primitive(Token {
+                        category: Data(vec![49]),
+                        position: (1, 29)
+                    })]
                 )),
                 Some(Box::new(Call(
                     Token {
-                        category: Identifier(None),
-                        position: (39, 46)
+                        category: Identifier(Undefined("display".to_owned())),
+                        position: (1, 40)
                     },
-                    Box::new(Parameter(vec![Primitive(Token {
-                        category: String(StringCategory::Quotable),
-                        position: (48, 50)
-                    })]))
+                    vec![Primitive(Token {
+                        category: Data(vec![104, 105]),
+                        position: (1, 48)
+                    })]
                 )))
             )
         );
@@ -389,8 +456,8 @@ mod test {
             actual,
             If(
                 Box::new(Variable(Token {
-                    category: Identifier(None),
-                    position: (4, 15)
+                    category: Identifier(Undefined("description".to_owned())),
+                    position: (1, 5)
                 })),
                 Box::new(Block(vec![])),
                 None
@@ -405,16 +472,16 @@ mod test {
                 scope,
                 vec![
                     Variable(Token {
-                        category: Identifier(None),
-                        position: (10 + offset, 11 + offset),
+                        category: Identifier(Undefined("a".to_owned())),
+                        position: (1, 11 + offset),
                     }),
                     Variable(Token {
-                        category: Identifier(None),
-                        position: (13 + offset, 14 + offset),
+                        category: Identifier(Undefined("b".to_owned())),
+                        position: (1, 14 + offset),
                     }),
                     Variable(Token {
-                        category: Identifier(None),
-                        position: (16 + offset, 17 + offset),
+                        category: Identifier(Undefined("c".to_owned())),
+                        position: (1, 17 + offset),
                     }),
                 ],
             )
@@ -435,8 +502,8 @@ mod test {
         assert_eq!(
             parse("NULL;").next().unwrap().unwrap(),
             Primitive(Token {
-                category: Identifier(Some(Keyword::Null)),
-                position: (0, 4)
+                category: Identifier(IdentifierType::Null),
+                position: (1, 1)
             })
         );
     }
@@ -446,15 +513,15 @@ mod test {
         assert_eq!(
             parse("TRUE;").next().unwrap().unwrap(),
             Primitive(Token {
-                category: Identifier(Some(Keyword::True)),
-                position: (0, 4)
+                category: Identifier(IdentifierType::True),
+                position: (1, 1)
             })
         );
         assert_eq!(
             parse("FALSE;").next().unwrap().unwrap(),
             Primitive(Token {
-                category: Identifier(Some(Keyword::False)),
-                position: (0, 5)
+                category: Identifier(IdentifierType::False),
+                position: (1, 1)
             })
         );
     }
@@ -469,10 +536,7 @@ mod test {
         ];
         for call in test_cases {
             assert!(
-                matches!(
-                    parse(&format!("{};", call)).next().unwrap().unwrap(),
-                    Exit(_),
-                ),
+                matches!(parse(&format!("{call};")).next().unwrap().unwrap(), Exit(_),),
                 "{}",
                 call
             );
@@ -491,7 +555,7 @@ mod test {
         for call in test_cases {
             assert!(
                 matches!(
-                    parse(&format!("{};", call)).next().unwrap().unwrap(),
+                    parse(&format!("{call};")).next().unwrap().unwrap(),
                     Return(_),
                 ),
                 "{}",
@@ -503,6 +567,11 @@ mod test {
     #[test]
     fn for_loop() {
         let code = "for (i = 0; i < 10; i++) display('hi');";
+        assert!(matches!(
+            parse(code).next().unwrap().unwrap(),
+            For(_, _, _, _)
+        ));
+        let code = "for (i = 0; i < 10; ) i = 10;";
         assert!(matches!(
             parse(code).next().unwrap().unwrap(),
             For(_, _, _, _)
@@ -530,7 +599,7 @@ mod test {
         for call in test_cases {
             assert!(
                 matches!(
-                    parse(&format!("{};", call)).next().unwrap().unwrap(),
+                    parse(&format!("{call};")).next().unwrap().unwrap(),
                     ForEach(_, _, _),
                 ),
                 "{}",
@@ -556,16 +625,16 @@ mod test {
                 .unwrap(),
             FunctionDeclaration(
                 Token {
-                    category: Identifier(None),
-                    position: (9, 26)
+                    category: Identifier(Undefined("register_packages".to_owned())),
+                    position: (1, 10)
                 },
                 vec![Variable(Token {
-                    category: Identifier(None),
-                    position: (28, 31)
+                    category: Identifier(Undefined("buf".to_owned())),
+                    position: (1, 29)
                 })],
                 Box::new(Block(vec![Return(Box::new(Primitive(Token {
-                    category: Number(Base::Base10),
-                    position: (43, 44)
+                    category: Number(1),
+                    position: (1, 44)
                 })))]))
             )
         );
@@ -576,13 +645,13 @@ mod test {
                 .unwrap(),
             FunctionDeclaration(
                 Token {
-                    category: Identifier(None),
-                    position: (9, 26)
+                    category: Identifier(Undefined("register_packages".to_owned())),
+                    position: (1, 10)
                 },
                 vec![],
                 Box::new(Block(vec![Return(Box::new(Primitive(Token {
-                    category: Number(Base::Base10),
-                    position: (39, 40)
+                    category: Number(1),
+                    position: (1, 40)
                 })))]))
             )
         );
@@ -596,17 +665,17 @@ mod test {
                 Category::Equal,
                 AssignOrder::AssignReturn,
                 Box::new(Variable(Token {
-                    category: Category::Identifier(None),
-                    position: (0, 4)
+                    category: Category::Identifier(Undefined("arg1".to_owned())),
+                    position: (1, 1)
                 },)),
                 Box::new(Array(
                     Token {
-                        category: Category::Identifier(Some(Keyword::FCTAnonArgs)),
-                        position: (7, 21),
+                        category: Category::Identifier(IdentifierType::FCTAnonArgs),
+                        position: (1, 8),
                     },
                     Some(Box::new(Primitive(Token {
-                        category: Category::Number(Base::Base10),
-                        position: (22, 23)
+                        category: Category::Number(0),
+                        position: (1, 23)
                     })))
                 ))
             ))
@@ -617,13 +686,13 @@ mod test {
                 Category::Equal,
                 AssignOrder::AssignReturn,
                 Box::new(Variable(Token {
-                    category: Category::Identifier(None),
-                    position: (0, 4)
+                    category: Category::Identifier(Undefined("arg1".to_owned())),
+                    position: (1, 1)
                 },)),
                 Box::new(Array(
                     Token {
-                        category: Category::Identifier(Some(Keyword::FCTAnonArgs)),
-                        position: (7, 21),
+                        category: Category::Identifier(IdentifierType::FCTAnonArgs),
+                        position: (1, 8),
                     },
                     None
                 ))
